"""Piprun main module."""

import os
import subprocess


def devnull():
    """Polyfill subprocess.DEVNULL for Python < 3.3."""
    try:
        return subprocess.DEVNULL
    except AttributeError:
        return open(os.devnull, 'w')


VIRTUALENV_DIR = os.path.expanduser('~/.piprun')


def split_args(args):
    """
    Split an argument list into arguments for piprun itself and the program
    being run. Everything before the first '--' is considered to be
    requirements, what follows '--' is passed to the program. It is an error
    not to have '--' in the list.
    """

    try:
        reqs_end = args.index('--')
    except ValueError:
        raise ValueError("Argument list must include '--'.")

    return args[:reqs_end], args[reqs_end + 1:]


def create_env(path, interpreter, requirements):
    """
    Create a virtual environment.

    :param path: Path of the new virtual environment
    :param interpreter: Python interpreter to use
    :param requirements: Requirements to install into the environment
    """

    virtualenv_args = ['virtualenv', '--quiet']
    if interpreter:
        virtualenv_args.append('--python={}'.format(interpreter))
    virtualenv_args.append(path)
    subprocess.check_call(virtualenv_args, stdout=devnull())

    pip = os.path.join(path, 'bin', 'pip')
    subprocess.check_call((pip, 'install') + requirements,
                          stdout=devnull())


def activate_env(path):
    """Set environment variables to activate the virtual environment."""

    os.environ['VIRTUAL_ENV'] = path

    old_path = os.environ.get('PATH', '')
    env_bin = os.path.join(path, 'bin')
    os.environ['PATH'] = env_bin + ':' + old_path


def main(*args):
    """Main entry point."""

    requirements, args = split_args(args)

    if os.path.exists(requirements[0]):
        interpreter = requirements[0]
        requirements = requirements[1:]
    else:
        interpreter = None

    requirements_hash = str(hash((interpreter, requirements)))

    env_path = os.path.join(VIRTUALENV_DIR, requirements_hash)

    if not os.path.exists(env_path):
        create_env(
            env_path,
            interpreter=interpreter,
            requirements=requirements,
        )

    activate_env(env_path)

    env_python = os.path.join(env_path, 'bin', 'python')
    os.execv(env_python, (env_python,) + args)



"""Test the testing functions."""

from __future__ import absolute_import
from __future__ import print_function

import sys
import os

from .base import ForkOutputTest


class TestForkOutputTest(ForkOutputTest):
    """Test the output test."""

    def test_fork_output(self):
        """Test fork_output."""

        self.assertEqual(
            self.fork_output(
                lambda: os.execl('/bin/echo', '/bin/echo', "asdf"),
            ),
            "asdf\n"
        )

    def test_returning_function(self):
        """Test a function that returns."""

        def returning_function():
            """Print some things and return."""
            print("asdf")
            sys.stdout.flush()

        self.assertEqual(
            self.fork_output(returning_function),
            "asdf\n"
        )

    def test_exception(self):
        """Test a function that raises an exception."""

        def raise_exception():
            """Raise a test exception."""
            raise Exception("asdf")

        self.assertIn(
            "Exception: asdf",
            self.fork_output(raise_exception),
        )



"""Base test classes."""

import os
import traceback
import unittest


class ForkOutputTest(unittest.TestCase):
    """Test functions that start new processes via exec()."""

    def fork_output(self, func):
        """
        Execute a given function in a subprocess and return the output.

        The function result is ignored.
        """

        out_read, out_write = os.pipe()

        if os.fork():
            # Parent process; read from the pipe
            # Close the write end of the pipe so the EOF is received when the
            # child exits
            os.close(out_write)
            with os.fdopen(out_read, 'r') as output:
                return output.read()
        else:
            # Child process; redirect standard output and defer to piprun
            # FIXME: Are stdout and stderr descriptors defined as constants
            # anywhere?
            os.dup2(out_write, 1)
            os.dup2(out_write, 2)
            try:
                func()
            except:  # pylint:disable=bare-except
                traceback.print_exc()
            os._exit(1)  # pylint:disable=protected-access



"""Test piprun."""

from __future__ import absolute_import

import shutil
import tempfile

from .base import ForkOutputTest

import piprun


class TestPiprun(ForkOutputTest):
    """Test piprun."""

    maxDiff = None

    def setUp(self):
        """
        Use a temporary directory for virtual environments generated by piprun.
        """

        virtualenv_dir = tempfile.mkdtemp()
        piprun.VIRTUALENV_DIR = virtualenv_dir
        self.addCleanup(shutil.rmtree, virtualenv_dir, ignore_errors=True)

    def execute_file(self, contents, *arguments):
        """
        Execute a file with specified contents using piprun.

        :param arguments: Arguments to pass to the process.
        :returns: Process output, as a string.
        """

        with tempfile.NamedTemporaryFile() as file_:
            file_.write(contents.encode())
            file_.flush()

            shebang = contents[:contents.find('\n')]
            self.assertEqual(shebang[:2], '#!',
                             "File must have a shebang.")
            shebang = shebang[2:].split()

            if shebang[0] == '/usr/bin/env':
                shebang = shebang[1:]

            self.assertEqual(shebang[0], 'piprun',
                             "File must be set up to run with piprun.")
            shebang = shebang[1:]

            arguments = shebang + [file_.name] + list(arguments)

            return self.fork_output(lambda: piprun.main(*arguments))

    FLASK_SCRIPT = """
#!/usr/bin/env piprun Flask==0.10.1 --
from __future__ import print_function

from flask import Flask

print(Flask)
""".strip()

    def test_execution(self):
        """Test executing a file that specifies dependencies."""

        self.assertEqual(
            self.execute_file(self.FLASK_SCRIPT),
            '''<class 'flask.app.Flask'>\n'''
        )
